# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
# Copyright (C) 2021 Enrico Ferreguti
# Copyright (C) 2012 Riccardo Apolloni
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''

import math
import inspect

from .. import utils

# http://www.tecepe.com.br/nav/vrtool/routing.htm

# Le isocrone saranno un albero; deve essere semplice:
# - accedere alla lista delle isocrone dell'ultimo albero
# - aggiungere un layer per il nuovo t
# - fare pruning di foglie

# [[level1], [level2,level2], [level3,level3,level3,level3]]


class RouterParam:
	def __init__(self, code, name, ttype, tooltip, default, lower=None, upper=None, step=None, digits=None):
		self.code = code
		self.name = name
		self.ttype = ttype
		self.tooltip = tooltip
		self.default = default
		self.value = default

		self.lower = lower
		self.upper = upper
		self.digits = digits
		self.step = step

class RoutingNoWindException(Exception):
	pass

class RoutingResult:
	def __init__(self, time, path=[], isochrones=[], position=None, progress=0):
		self.time = time
		self.path = path
		self.isochrones = isochrones
		self.position = position
		self.progress = progress

class Router:
	PARAMS = {}

	def __init__ (self, polar, grib, pointValidity = None, lineValidity = None):
		self.polar = polar
		self.grib = grib
		self.pointValidity = pointValidity
		self.lineValidity = lineValidity

	def setParamValue(self, code, value):
		self.PARAMS[code] = value

	def getParamValue(self, code):
		return self.PARAMS[code].value

	def calculateShortestPathIsochrones (self, fixedSpeed, t, isocrone, nextwp):
		""" Calculates isochrones based on shortest path at fixed speed (motoring);
			the speed considers reductions / increases derived from leeway """
		def pointF(p, tws, twa, dt, brg):
			# TODO: add current factor
			speed = fixedSpeed
			return utils.routagePointDistance (p[0], p[1], speed * dt * utils.NAUTICAL_MILE_IN_KM, brg), speed

		return self._calculateIsochrones(t, isocrone, nextwp, pointF)
		

	def calculateIsochrones (self, t, isocrone, nextwp):
		""" Calculate isochrones depending on routageSpeed from polar """
		def pointF(p, tws, twa, dt, brg):
			speed = self.polar.getSpeed (tws, math.copysign (twa,1))
			return utils.routagePointDistance (p[0], p[1], speed * dt * utils.NAUTICAL_MILE_IN_KM, brg), speed

		return self._calculateIsochrones(t, isocrone, nextwp, pointF)


	def _calculateIsochrones (self, t, isocrone, nextwp, pointF):
		""" Calcuates isochrones based on pointF next point calculation """
		dt = (1. / 60. * 60.)
		last = isocrone [-1]

		newisopoints = []

		# foreach point of the iso
		for i in range (0, len (last)):
			p = last[i]

			try:
				(twd,tws) = self.grib.getWindAt (t, p[0], p[1])
			except:
				raise (RoutingNoWindException())

			for twa in range(-180,180,5):
				twa = math.radians(twa)
				brg = utils.reduce360(twd + twa)

				# Calculate next point
				ptoiso, speed = pointF(p, tws, twa, dt, brg)
				
				if utils.pointDistance (ptoiso[0], ptoiso[1], nextwp[0], nextwp[1]) >= utils.pointDistance (p[0], p[1], nextwp[0], nextwp[1]):
				 	continue
				
				if self.pointValidity:
					if not self.pointValidity (ptoiso[0], ptoiso[1]):
						continue
				else:
					if not self.lineValidity (ptoiso[0], ptoiso[1], p[0], p[1]):
						continue
				
				newisopoints.append ((ptoiso[0], ptoiso[1], i, t, twd, tws, speed, math.degrees(brg)))


		# sort newisopoints based on bearing
		isonew = []
		for i in range(0, len (newisopoints)):
			try:
				newisopoints[i] = (newisopoints[i][0], newisopoints[i][1], newisopoints[i][2], utils.lossodromic (isocrone[0][0][0],isocrone[0][0][1],newisopoints[i][0],newisopoints[i][1]), newisopoints[i][3], newisopoints[i][4], newisopoints[i][5], newisopoints[i][6], newisopoints[i][7])
				isonew.append (newisopoints[i])
			except:
				pass

		newisopoints = sorted (isonew, key=(lambda a: a[3][1]))

		# remove slow isopoints inside
		bearing = {}
		for x in newisopoints:
			k = str (int (x[3][1] * 100))
			if k in bearing:
				if x[3][0] > bearing[k][3][0]:
					bearing[k] = x
			else:
				bearing[k] = x

		isonew = []
		for x in bearing:	
			isonew.append (bearing[x])


		isonew = sorted (isonew, key=(lambda a: a[3][1]))
		isocrone.append (isonew)

		return isocrone


	def calculateVMG (self, speed, angle, start, end) -> float:
		""" Calculates the Velocity-Made-Good of a boat sailing from start to end at current speed / angle """
		return speed * math.cos (angle)


	def route (self, lastlog, t, start, end) -> RoutingResult:
		raise (Exception("Not implemented"))



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
# Copyright (C) 2021 Enrico Ferreguti
# Copyright (C) 2012 Riccardo Apolloni
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''

import latlon
from typing import Tuple 
import math

EARTH_RADIUS = 60.0 * 360 / (2 * math.pi) # nm
NAUTICAL_MILE_IN_KM = 1.852

def cfbinomiale(n: float, i: float) -> float:
	return math.factorial(n)/(math.factorial(n-i)*math.factorial(i))

def ortodromic2 (lat1: float, lon1: float, lat2: float, lon2: float) -> Tuple[float, float]:
	p1 = math.radians (lat1)
	p2 = math.radians (lat2)
	dp = math.radians (lat2-lat1)
	dp2 = math.radians (lon2-lon1)

	a = math.sin (dp/2) * math.sin (dp2/2) + math.cos (p1) * math.cos (p2) * math.sin (dp2/2) * math.sin (dp2/2)
	c = 2 * math.atan2(math.sqrt(a), math.sqrt(1-a))
	return (EARTH_RADIUS * c, a)

def ortodromic (latA: float, lonA: float, latB: float, lonB: float) -> Tuple[float, float]:
	p1 = latlon.LatLon(latlon.Latitude(latA), latlon.Longitude(lonA))
	p2 = latlon.LatLon(latlon.Latitude(latB), latlon.Longitude(lonB))

	return (p1.distance (p2), math.radians (p1.heading_initial(p2)))

def lossodromic (latA: float, lonA: float, latB: float, lonB: float) -> Tuple[float, float]:
	p1 = latlon.LatLon(latlon.Latitude(latA), latlon.Longitude(lonA))
	p2 = latlon.LatLon(latlon.Latitude(latB), latlon.Longitude(lonB))

	return (p1.distance (p2, ellipse = 'sphere'), math.radians (p1.heading_initial(p2)))


def km2nm(d: float) -> float:
	return d * 0.539957

def nm2km(d: float) -> float:
	return d / 0.539957

def pointDistance (latA: float, lonA: float, latB: float, lonB: float, unit: str = 'nm') -> float:
	""" Returns the distance between two geo points """
	p1 = latlon.LatLon(latlon.Latitude(latA), latlon.Longitude(lonA))
	p2 = latlon.LatLon(latlon.Latitude(latB), latlon.Longitude(lonB))
	d = p1.distance (p2)

	if unit == 'nm':
		return km2nm(d)
	elif unit == 'km':
		return d
	

def routagePointDistance (latA: float, lonA: float, distance: float, hdg: float, unit: str='nm') -> Tuple[float, float]:
	""" Returns the point from (latA, lonA) to the given (distance, hdg) """
	if unit == 'nm':
		d = nm2km(distance)
	elif unit == 'km':
		d = distance

	p = latlon.LatLon(latlon.Latitude(latA), latlon.Longitude(lonA))
	of = p.offset (math.degrees (hdg), d).to_string('D')
	return (float (of[0]), float (of[1]))


def maxReachDistance(p, speed: float, dt: float = (1. / 60. * 60.)) -> float:
	maxp = routagePointDistance (p[0], p[1], speed * dt, 1)
	return pointDistance(p[0], p[1], maxp[0], maxp[1])


def reduce360 (alfa: float) -> float:
	if math.isnan (alfa):
		return 0.0
		
	n=int(alfa*0.5/math.pi)
	n=math.copysign(n,1)
	if alfa>2.0*math.pi:
		alfa=alfa-n*2.0*math.pi
	if alfa<0:
		alfa=(n+1)*2.0*math.pi+alfa
	if alfa>2.0*math.pi or alfa<0:
		return 0.0
	return alfa

def reduce180 (alfa: float) -> float:
	if alfa>math.pi:
		alfa=alfa-2*math.pi
	if alfa<-math.pi:
		alfa=2*math.pi+alfa
	if alfa>math.pi or alfa<-math.pi:
		return 0.0
	return alfa

def pathAsGeojson(path: str):
	tr = []
	for wp in path:
		if len(wp) == 3:
			tr.append((wp[0], wp[1], str(wp[2]), 0, 0, 0, 0))
		else:
			tr.append((wp[0], wp[1], str(wp[4]), wp[5], wp[6], wp[7], wp[8]))

	feats = []
	route = []

	for order,wayp in enumerate(tr):
		feat = {
			"type": "Feature",
			"id": order,
			"geometry": {
				"type": "Point",
				"coordinates": [ # longitude, latitude
					wayp[1],
					wayp[0]
				]
			},
			"properties": {
				"timestamp": str(wayp[2]),
				"twd": math.degrees(wayp[3]),
				"tws": wayp[4],
				"knots": wayp[5],
				"heading": wayp[6]
			}
		}
		feats.append(feat)
		route.append([wayp[1], wayp[0]]) # longitude, latitude

	feats.append({
		"type": "Feature",
		"id": order+1,
		"geometry": {
			"type": "LineString",
			"coordinates": route
		},
		"properties": {
			"start-timestamp": str(tr[0][2]),
			"end-timestamp": str(tr[-1][2])
		}
	})

	gj = {
		"type": "FeatureCollection",
		"features": feats
	}

	return gj



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''

from .routers import linearbestisorouter, RoutingResult

def listRoutingAlgorithms():
	""" Return a list of routing algorithms along with their names"""

	return [
		{
			'name': 'LinearBestIsoRouter',
			'class': linearbestisorouter.LinearBestIsoRouter
		}
	]

class Routing:
	"""
	"""

	def __init__ (self, algorithm, polar, track, grib, startDatetime, startPosition = None, pointValidity = None, lineValidity = None):
		"""
		Parameters
		----------
		algorithm : Router
			The routing algorithm class
		polar : Polar
			Polar object of the boat we want to route
		track : list
			A list of track points (lat, lon)
		grib : Grib
			Grib object that abstract our wind / wave / wathever queries
		startDateTime : datetime
			Start time
		startPosition : (float, float)
			Optional, default to None
			A pair containing initial position (or None if we want to start from the first track point)
		pointValidity : function(lat, lon)
			Optional, default to None
			A functions that receives lat and lon and returns True if the point is valid (ie: in the sea)
		lineValidity : function(lat1, lon1, lat2, lon2)
			Optional, default to None
			A functions that receives a vector defined by lat1, lon1, lat2, lon2 and returns True if the line is valid (ie: completely in the sea)
		"""

		self.end = False
		self.algorithm = algorithm(polar, grib, pointValidity, lineValidity)
		self.track = track
		self.steps = 0
		self.path = []
		self.time = startDatetime
		self.grib = grib
		self.log = []   

		if startPosition:
			self.wp = 0
			self.position = startPosition
		else:
			self.wp = 1
			self.position = self.track[0]


	def step (self) -> RoutingResult:
		""" Execute a single routing step """

		self.steps += 1
		
		if self.wp >= len (self.track):
			self.end = True
			res = self.log[-1]
			return self.log[-1]

		# Next waypoint
		nextwp = self.track[self.wp]

		if len (self.log) > 0:
			res = self.algorithm.route (self.log[-1], self.time, self.position, nextwp)
		else:
			res = self.algorithm.route (None, self.time, self.position, nextwp)


		#self.time += 0.2
		progress = len (self.log) * 5

		if len (res.path) != 0:
			self.position = res.position
			self.path = self.path + res.path
			self.wp += 1
			res.isochrones = []

		np = []
		ptime = None 
		for x in self.path:
			if len(x) > 3: nt = x[4]
			else: nt = x[2]

			if ptime:
				if ptime < nt:
					np.append(x)
					ptime = nt
			else:
				np.append(x)
				ptime = nt

		self.path = np
		self.time = res.time
		nlog = RoutingResult(progress=progress, time=res.time, path=self.path, isochrones=res.isochrones)

		self.log.append (nlog)
		return nlog
		


# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
# Copyright (C) 2021 Enrico Ferreguti
# Copyright (C) 2012 Riccardo Apolloni
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''

import datetime
from .. import utils
from .linearbestisorouter import *

class ShortestPathRouter (LinearBestIsoRouter):
	PARAMS = {
		'minIncrease': RouterParam('minIncrease', 'Minimum increase (nm)', 'float', 'Set the minimum value for selecting a new valid point', default=10.0, lower=1.0, upper=100.0, step=0.1, digits=1),
		'fixedSpeed': RouterParam('fixedSpeed', 'Fixed speed (kn)', 'float', 'Set the fixed speed', default=5.0, lower=1.0, upper=60.0, step=0.1, digits=1)
	}

	def route (self, lastlog, time, start, end) -> RoutingResult:
		return self._route(lastlog, time, start, end, lambda time, isoc, end: self.calculateShortestPathIsochrones (self.getParamValue('fixedSpeed'), time, isoc, end))



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''
from .grib import Grib
from .polar import Polar
from .routing import Routing, listRoutingAlgorithms
from .utils import *
from .routers import *


# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''
from typing import Tuple 

class Grib:
	""" 
	Grib class is an abstract class that should be implement for providing grib data to routers
	"""

	def getWindAt(self, t: float, lat: float, lon: float) -> Tuple[float, float]:
		""" 
		Returns (twd: radians, tws: knots) for the given point (lat, lon) at time t 
		or None if running out of temporal/geographic grib scope
		"""
		raise (Exception('Not implemented'))




# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''

from setuptools import find_packages
from setuptools import setup

setup(name='weatherrouting',
	version='0.1.6',
	description='Weather routing library for sailing',
	author=['Davide Gessa'],
	setup_requires='setuptools',
	author_email='gessadavide@gmail.com',
	packages=['weatherrouting', 'weatherrouting.routers'],
	install_requires=['latlon3'], 
	test_suite = 'tests',
)



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
# Copyright (C) 2021 Enrico Ferreguti
# Copyright (C) 2012 Riccardo Apolloni
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''
from typing import Tuple 
from io import FileIO
import math

class Polar:
    def __init__ (self, polarPath: str, f: FileIO = None):
        """
        Parameters
        ----------
        polarPath : string
            Path of the polar file
        f : File
            File object for passing an opened file
        """

        self.tws = []
        self.twa = []
        self.vmgdict = {}
        self.speedTable = []
        
        if f == None:
            f = open (polarPath, "r")
        
        tws = f.readline ().split ()
        for i in range (1,len(tws)):
            self.tws.append (float (tws[i].replace ('\x02', '')))

        line = f.readline ()
        while line != "":
            data = line.split ()
            twa = float (data[0])
            self.twa.append (math.radians (twa))
            speedline = []
            for i in range (1,len (data)):
                speed = float (data[i])
                speedline.append (speed)
            self.speedTable.append (speedline)
            line = f.readline ()
        f.close ()


    def toString(self) -> str:
        s = 'TWA\\TWS'
        for x in self.tws:
            s += '\t%d' % x
        s += '\n'

        l = 0
        for y in self.twa:
            s += '%d' % round(math.degrees(y))
            sl = self.speedTable[l]

            for x in sl:
                s += '\t{:.1f}'.format(x)
            s += '\n'

            l += 1

        return s
        
    def getSpeed (self, tws: float, twa: float) -> float:
        """ Returns the speed (in knots) given tws (in knots) and twa (in radians) """
        
        tws1 = 0
        tws2 = 0
        for k in range(0, len(self.tws)):
            if tws >= self.tws[k]:
               tws1 = k
        for k in range(len(self.tws) - 1, 0, -1):
            if tws <= self.tws[k]:
                tws2 = k
        if tws1 > tws2: # TWS over table limits
            tws2 = len(self.tws) - 1
        twa1 = 0
        twa2 = 0
        for k in range(0, len(self.twa)):
            if twa >= self.twa[k]:
               twa1 = k
        for k in range(len(self.twa) - 1, 0, -1):
            if twa <= self.twa[k]:
                twa2 = k

        speed1 = self.speedTable[twa1][tws1]
        speed2 = self.speedTable[twa2][tws1]
        speed3 = self.speedTable[twa1][tws2]
        speed4 = self.speedTable[twa2][tws2]

        if twa1 != twa2:
            speed12 = speed1 + (speed2 - speed1) * (twa-self.twa[twa1]) / (self.twa[twa2] - self.twa[twa1])#interpolazione su twa
            speed34 = speed3 + (speed4 - speed3) * (twa-self.twa[twa1]) / (self.twa[twa2] - self.twa[twa1])#interpolazione su twa
        else:
            speed12 = speed1
            speed34 = speed3
        if tws1 != tws2:
            speed = speed12 + (speed34 - speed12) * (tws - self.tws[tws1]) / (self.tws[tws2] - self.tws[tws1])
        else:
            speed = speed12
        return speed
        

    def getReaching(self, tws: float) -> Tuple[float, float]:
        maxspeed = 0
        twamaxspeed = 0
        for twa in range(0,181):
            twa = math.radians(twa)
            speed = self.getSpeed (tws,twa)
            if speed > maxspeed:
                maxspeed = speed
                twamaxspeed = twa
        return (maxspeed, twamaxspeed)


    def getMaxVMGTWA(self, tws: float, twa: float) -> Tuple[float, float]:
        if not ((tws, twa) in self.vmgdict):
            twamin = max(0, twa-math.pi/2)
            twamax = min(math.pi, twa+math.pi/2)
            alfa = twamin
            maxvmg = -1.0
            while alfa < twamax:
                v = self.getSpeed (tws, alfa)
                vmg = v * math.cos(alfa-twa)
                if vmg - maxvmg > 10**-3:#10**-3 errore tollerato 
                    maxvmg = vmg
                    twamaxvmg = alfa
                alfa = alfa + math.radians(1)
            self.vmgdict[tws, twa] = (maxvmg, twamaxvmg)
        return self.vmgdict[(tws, twa)]


    def getMaxVMGUp(self, tws: float) -> Tuple[float, float]:
        vmguptupla = self.getMaxVMGTWA(tws, 0)
        return (vmguptupla[0], vmguptupla[1])


    def getMaxVMGDown(self, tws: float) -> Tuple[float, float]:
        vmgdowntupla = self.getMaxVMGTWA(tws, math.pi)
        return (-vmgdowntupla[0], vmgdowntupla[1])


    def getRoutageSpeed(self, tws, twa) -> float:
        UP = self.getMaxVMGUp(tws)
        vmgup = UP[0]
        twaup = UP[1]
        DOWN = self.getMaxVMGDown(tws)
        vmgdown = DOWN[0]
        twadown = DOWN[1]
        v = 0.0

        if twa >= twaup and twa <= twadown:
            v = self.getSpeed (tws, twa)
        else:
            if twa < twaup:
                v = vmgup / math.cos(twa)
            if twa > twadown:
                v = vmgdown / math.cos(twa)
        return v


    def getTWARoutage(self, tws: float, twa: float) -> float:
        UP = self.getMaxVMGUp(tws)
        vmgup = UP[0]
        twaup = UP[1]
        DOWN = self.getMaxVMGDown(tws)
        vmgdown = DOWN[0]
        twadown = DOWN[1]
        if twa >= twaup and twa <= twadown:
            twa = twa
        else:
            if twa < twaup:
                twa = twaup
            if twa > twadown:
                twa = twadown
        return twa

        


# -*- coding: utf-8 -*-
# Copyright (C) 2017-2021 Davide Gessa
# Copyright (C) 2021 Enrico Ferreguti
# Copyright (C) 2012 Riccardo Apolloni
'''
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For detail about GNU see <http://www.gnu.org/licenses/>.
'''

import datetime
from .. import utils
from .router import *

class LinearBestIsoRouter (Router):
	PARAMS = {
		'minIncrease': RouterParam('minIncrease', 'Minimum increase (nm)', 'float', 'Set the minimum value for selecting a new valid point', default=10.0, lower=1.0, upper=100.0, step=0.1, digits=1)
	}


	def _route (self, lastlog, time, start, end, isoF):
		position = start
		path = []

		def generate_path(p):
			nonlocal path
			nonlocal isoc
			nonlocal position
			path.append (p)
			for iso in isoc[::-1][1::]:
				path.append (iso[path[-1][2]])
			path = path[::-1]
			position = path[-1]
		
		if self.grib.getWindAt (time + datetime.timedelta(hours=1), end[0],end[1]):
			if lastlog != None and len (lastlog.isochrones) > 0:
				isoc = isoF(time + datetime.timedelta(hours=1), lastlog.isochrones, end)
			else:
				isoc = isoF(time + datetime.timedelta(hours=1), [[(start[0], start[1], time)]], end)

			nearest_dist = self.getParamValue('minIncrease')
			nearest_solution = None
			for p in isoc[-1]:
				distance_to_end_point = utils.pointDistance (end[0],end[1], p[0], p[1])
				if distance_to_end_point < self.getParamValue('minIncrease'):
					(twd,tws) = self.grib.getWindAt (time + datetime.timedelta(hours=1), p[0], p[1])
					maxReachDistance = utils.maxReachDistance(p, p[6])
					if utils.pointDistance (end[0],end[1], p[0], p[1]) < abs(maxReachDistance*1.1):
						if (not self.pointValidity or self.pointValidity(end[0],end[1])) and (not self.lineValidity or self.lineValidity(end[0],end[1], p[0], p[1])):
							if distance_to_end_point < nearest_dist:
								nearest_dist = distance_to_end_point
								nearest_solution = p
			if nearest_solution:
				generate_path(nearest_solution)

		# out of grib scope
		else: 
			minDist = 1000000
			isoc = lastlog.isochrones
			for p in isoc[-1]:
				checkDist = utils.pointDistance (end[0],end[1], p[0], p[1]) 
				if checkDist < minDist:
					minDist = checkDist
					minP = p
			generate_path(minP)

		return RoutingResult(time=time + datetime.timedelta(hours=1), path=path, position=position, isochrones=isoc)


	def route (self, lastlog, time, start, end) -> RoutingResult:
		return self._route(lastlog, time, start, end, self.calculateIsochrones)
